#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-05-31 09:41:05.625019"
__version__ = "@COSTO_VERSION@"
# **************************************************
from . import couplingData as CODA
import mpi4py.MPI as MPI
import numpy as np


def init_MPI_and_build_local_communicator(c_data:CODA.couplingData=None):
    if not MPI.Is_initialized():
        print("MPI init must be called before")
    # ATTENTION Dup => collective comms
    # c_data.comm_w = MPI.COMM_WORLD.Dup()
    c_data.comm_w = MPI.COMM_WORLD
    c_data.comm_l = c_data.comm_w.Split(c_data.color, key=0)
    c_data.rank_w = c_data.comm_w.Get_rank()
    c_data.size_w = c_data.comm_w.Get_size()
    c_data.rank_l = c_data.comm_l.Get_rank()
    c_data.size_l = c_data.comm_l.Get_size()
    c_data.is_initialize = True
    c_data.print("(global) I m {}/{}".format(c_data.rank_w+1, c_data.size_w))
    c_data.print("(local) I m {}/{}".format(c_data.rank_l+1, c_data.size_l))
    return


def recv_data(comm, src=0, tag=0):
    shape = np.zeros(3, dtype=np.int32)
    comm.Recv(shape, source=src, tag=tag)
    size = shape[0]*shape[1]+shape[2]
    data = np.zeros(size, dtype=np.double)
    comm.Recv(data, source=src, tag=tag+1)
    return data, shape


def send_data(comm, data, shape, dest=0, tag=0):
    comm.Send(shape, dest=dest, tag=tag)
    comm.Send(data, dest=dest, tag=tag+1)
    return


def finalize_MPI():
    MPI.Finalize()
# def recv_points(src=0, tag=100):
#     data, shape = recv_data(src, tag)
#     assert(shape[2] == 0)
#     data.shape = (shape[0], shape[1])

#     pts_IFS_solid = np.zeros((shape[0], 3))
#     for i in range(shape[1]):
#         pts_IFS_solid[:, i] = data[:, i]
#     return pts_IFS_solid



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-07-03 19:08:08.435410"
__version__ = "@COSTO_VERSION@"
# **************************************************
import mpi4py.MPI as MPI
import numpy as np
import inspect
from .neibor import *

def do_nothing():
    return

class IBC():
    def __init__(self, *args, name: str = "", trace: np.array = None, **kwds):
        self.name = name
        self._trace = trace

    def _apply_header(self):
        print("apply BC {} (type: {})".format(self.name,
                                              self.__class__.__name__))
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = do_nothing

    def compute_pre_vr(self):
        """
        interface methode
        """
        raise NotImplementedError("Must be override")


    def apply_pre_vr(self):
        """
        apply the boundary condition
        """
        self._apply_header()
        self.compute_pre_vr()

    def compute_post_vr(self):
        """
        interface methode
        """
        return

    def apply_post_vr(self):
        """
        apply the boundary condition
        """
        self.compute_post_vr()

    def __repr__(self):
        to_ret = "-"*50
        to_ret += "\nBC name: " + self.name
        # to_ret += "\nrank: " + str(self._rank)
        # to_ret += "\nneibors: " + str(self.neibors_ranks)
        to_ret += "\n"+"-"*50
        return to_ret


class IInternal_BC():
    def __init__(self, *args, comm: MPI.Comm = MPI.COMM_NULL, neibors: list = None, verbose: int = 1, dim=1, nr=0, **kwds):
        self._comm = comm
        self.neibor_ranks = neibors           # rang MPI dans le comm des voisins
        self._nb_neibors = len(neibors)
        self.dim = dim
        self._nr = nr
        self.Ar = None
        self.Br = None
        self._buff_send = None
        self._buff_recv = None
        self.neibors: list(IInternalNeibor) = []
        self._allocate_arrays()
        self._insert_neibors()

        if verbose > 0:
            print(self.__repr__())
        return

    def _allocate_arrays(self):
        self.Ar = np.zeros((self._nr, self.dim, self.dim), dtype=np.double)
        self.Br = np.zeros((self._nr, self.dim), dtype=np.double)
        self._dimAr = self._nr*self.dim**2
        self._dimBr = self._nr*self.dim
        self._dimArpBr = self._dimAr+self._dimBr
        self._buff_dim_by_neibor = self._nr*(self.dim+self.dim**2) # taille de Ar + Br
        self._buff_send = np.empty((self._buff_dim_by_neibor), dtype=np.double)
        self._buff_recv = np.empty((self._nb_neibors, self._buff_dim_by_neibor), dtype=np.double)

    def _insert_neibors(self):
        for i, rank in enumerate(self.neibor_ranks):
            intNeibor = IInternalNeibor(comm=self._comm,
                                       rank=rank,
                                       # self._nr,
                                       # selfelf.dim,
                                       buff_send=self._buff_send,
                                       buff_recv=self._buff_recv[i])
            self.neibors.append(intNeibor)

    def _publish_data(self):
        for neibor in self.neibors:
            neibor.publish()

    def _collect_data(self):
        for neibor in self.neibors:
            neibor.collect()

    def _update_Ar_Br(self):
        self.cumNeibors = np.sum(self._buff_recv, axis=0)
        self.Br += self.cumNeibors[:self._dimBr].reshape(self._nr, self.dim)
        self.Ar += self.cumNeibors[self._dimBr: self._dimArpBr].reshape(self._nr, self.dim, self.dim)

    def fill_Ar_and_Br(self):
        # self.Ar[:] = self.solver.Ar[self._trace]
        # self.Br[:] = self.solver.Br[self._trace]
        raise NotImplementedError("Must Be ovverride")

    def _extract_data(self):
        self.fill_Ar_and_Br()
        self._buff_send[: self._dimBr] = self.Br.flatten()
        self._buff_send[self._dimBr:self._dimArpBr] = self.Ar.flatten()

    def _do_exchanges(self):
        self._publish_data()     # envoi son Ajr (j = self) à qui de droit (j != 1)
        self._collect_data()  # recoi les Ajr (j != self)
        self._update_Ar_Br()  # somme sur tout les Ajr sur le neoud r

    def set_new_data(self):
        # self.solver.Br[self._trace] = self.Br
        # self.solver.Ar[self._trace] = self.Ar
        raise NotImplementedError("Must Be ovverride")

    def compute_pre_vr(self):
        self._extract_data() # get local Ar and Br and fill the sendding buffer
        self._do_exchanges() # update Ar and Br from neibors and send mines
        self.set_new_data() # Apply the new Ar and Br



class ICoupled_BC(IInternal_BC):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        return

    def _insert_neibors(self):
        for i, rank in enumerate(self.neibor_ranks):
            neibor = INeibor(comm=self._comm,
                             rank=rank,
                             # self._nr,
                             # self.dim,
                             buff_send=self._buff_send,
                             buff_recv=self._buff_recv[i])
            self.neibors.append(neibor)

    def set_new_data(self):
        # self.solver.Br[self._trace] = self.Br
        # self.solver.Ar[self._trace] = self.Ar
        raise NotImplementedError("Must Be ovverride")


    def fill_Ar_and_Br(self):
        # self.Ar[:] = self.solver.Ar[self._trace]
        # self.Br[:] = self.solver.Br[self._trace]
        raise NotImplementedError("Must Be ovverride")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-06-28 23:09:27.064258"
__version__ = "@COSTO_VERSION@"
# **************************************************
from . import utils as COUT


class couplingData():
    def __init__(self, code_name):
        __slots__ = ("code_name","color","printer","print","is_initialize","rank_l",
                     "rank_w","size_l","size_w","comm_l","comm_w")
        self.code_name = code_name
        self.color = sum([ord(c) for c in self.code_name])
        self.printer = COUT.printer(mention="({})".format(self.code_name))
        self.print = self.printer.print
        self.is_initialize = False
        self.rank_l = -1
        self.rank_w = -1
        self.size_l = -1
        self.size_w = -1
        self.comm_l = None
        self.comm_w = None

    def __repr__(self):
        to_ret = ""
        to_ret += "code name: {}\n".format(self.code_name)
        to_ret += "associated ID: {}\n".format(self.color)
        if self.is_initialize:
            to_ret += "rank world: {}\n".format(self.rank_w)
            to_ret += "rank local: {}\n".format(self.rank_l)
            to_ret += "size world communicator: {}\n".format(self.size_w)
            to_ret += "size local communicator: {}\n".format(self.size_l)
        return to_ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-05-31 12:18:14.567090"
__version__ = "@COSTO_VERSION@"
# **************************************************
from . import utils as COUT
from . import pipeline as COPI
from . import connectors as COCO
from . import timeManager as COTI
from . import couplingData as CODA
from .clock import clock

import numpy as np
import inspect


class coupling(COUT.utils, ):

    def __init__(self, code_name="codex" ):
        super().__init__()
        self.code_name = code_name
        self.c_data = CODA.couplingData(self.code_name)
        self.printer = self.c_data.printer
        self.mesh = None
        self.connector_type = None
        self.connector:  COCO.Iconnector = None
        self.other_code_ranks = None
        self.others_code_ID = None
        self.same_code_ranks = None
        self.timeStepComputer = COTI.TimeStepComputer(self)
        self.clock = clock(0.)

    def initialize(self):
        COPI.init_MPI_and_build_local_communicator(self.c_data)
        codes_ID = np.ones(self.c_data.size_w, dtype=np.int32)*-1
        color = np.ones(1, dtype=np.int32)*self.c_data.color
        self.c_data.comm_w.Allgather(color, codes_ID)
        self.other_code_ranks = np.argwhere(codes_ID != self.c_data.color).flatten()
        self.other_code_ranks.sort()
        self.others_code_ID = np.array(codes_ID[self.other_code_ranks])
        self.same_code_ranks = np.argwhere(codes_ID == self.c_data.color).flatten()
        self.same_code_ranks.sort()

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def set_connector_type(self, connector_type):
        self.connector_type = connector_type
        assert(self.connector_type != None)

    def set_mesh(self, mesh):
        if self.connector_type is None:
            raise NotImplementedError("You must set your a connector type before calling set_mesh")
        self.mesh = mesh
        assert(self.mesh != None)
        self.connector = self.connector_type(univers=self.c_data.comm_w, mesh=self.mesh)

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def build_graph(self, *args, **kwds):
        if self.connector is None:
            raise NotImplementedError("You must set your mesh before calling build_graph")
        self.connector.compute_graph(*args, **kwds)

    def build_graph_based_on_skin_points(self, mesh):
        self.set_connector_type(COCO.connect_skin_points)
        self.set_mesh(mesh)
        self.build_graph(fields_racine="graph", add_individual_graphs=False, build_nodal_graph=True)
        res = self.connector.get_traces_and_map()
        return res

    def build_graph_based_on_cloud_points(self, mesh):
        self.set_connector_type(COCO.connect_points)
        self.set_mesh(mesh)
        self.build_graph(fields_racine="graph", add_individual_graphs=False, build_nodal_graph=True)
        res = self.connector.get_traces_and_map()
        return res

    def finalize(self):
        COPI.finalize_MPI()

    def __repr__(self):
        to_ret = ""
        to_ret += self.c_data.__repr__()
        return to_ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-11-09 17:39:32.284214"
__version__ = 1.0

import costo.utils as  COUL
import os
import pyvista as pv
import numpy as np
import traceback

src_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))

class test_class_A():

    def __init__(self):
        self.a = 0

class test_class_B(test_class_A):

    def __init__(self):
        self.b = 0

class test_class_C(test_class_B):

    def __init__(self):
        self.c = 0

def read(fn, verbose=False):
    full_fn = os.path.join(src_dir, fn)
    if verbose:
        print(full_fn)
    return pv.read(full_fn)

def test_check_input():
    dic = {"k1": "a",
           "k2": "a"}
    COUL.check_inputs(dic, "k1")
    COUL.check_inputs(dic, "k2")
    try:
        COUL.check_inputs(dic, "k3")
    except:
        print("undifined key")
    return

def test_class_printer():
    p = COUL.printer()
    p.print("COUCOU BLUE")

    p.color  = p.Fore.GREEN
    p.print("COUCOU GREEN")
    return

def test_class_utils():
    class test_utils(COUL.utils):
        def __init__(self):
            return

    ult = COUL.utils()

    c = test_utils()
    c.do_nothing()
    c.do_nothing(1., 20, True)

    def fct_test():
        print("call fct test")

    f = test_utils.call_only_once(fct_test)
    f()
    return


def test_FindAllSubClasses():
    res = COUL.FindAllSubclasses(test_class_A)
    print(res)
    className = list(ires[1] for ires in res)
    className.sort()
    assert className == ['test_class_B', 'test_class_C']

def test_compare_mesh():
    m = read("00-DATA/first_order_elements.vtu", verbose=True)
    mp1 = pv.PolyData([[0, 0., 0.]])
    mp1.point_data["a"] = [0]
    mp1.cell_data["a"] = [0]
    COUL.compare_mesh(mp1, mp1)

    mp2 = pv.PolyData([[0, 0., 0.],
                       [1.,0.,0.]])

    mp3 = pv.PolyData([[1e-5, 0., 0.]])
    mp3.point_data["a"] = [0]
    mp3.cell_data["a"] = [0]

    mp4 = pv.PolyData([[0., 0., 0.]])
    mp4.cell_data["a"] = [0]

    mp7 = pv.PolyData([[0., 0., 0.]])
    mp7.point_data["a"] = [0]

    try:
        COUL.compare_mesh(m, mp1)
    except:
        print("Try to compare different mesh types")


    try:
        COUL.compare_mesh(None, None)
    except:
        print("Mesh type out of availlable mesh types")

    try:
        COUL.compare_mesh(mp1, mp2)
    except:
        print("Number of point are differents")

    try:
        COUL.compare_mesh(mp1, mp3)
    except:
        print("Try to compare meshs with differents Points")

    try:
        COUL.compare_mesh(mp1, mp4)
    except:
        print("Try to compare meshs with different number of point_data")


    mp5 = pv.PolyData([[0., 0., 0.],
                       [1., 0., 0.],
                       [2., 0., 0.]],lines=[2,0,1,
                                            2,1,2]
                      )
    mp6 = pv.PolyData([[0., 0., 0.],
                       [1., 0., 0.],
                       [2., 0., 0.]],lines=[2,0,1,
                                            2,1,2,
                                            2,0,2]
                      )

    try:
        COUL.compare_mesh(mp5, mp6)
    except:
        print("Try to compare mesh polydata with different number of cell")

    try:
        COUL.compare_mesh(mp1, mp7)
    except:
        print("Différent number of cell_data")


    cells = np.array([[4, 0, 1, 2, 3],
                      [4, 4, 5, 6, 7]])
    celltypes = [pv.CellType.TETRA,
                 pv.CellType.TETRA]
    points = np.array([
        [1.0, 1.0, 1.0],
        [1.0, -1.0, -1.0],
        [-1.0, 1.0, -1.0],
        [-1.0, -1.0, 1.0],
        [2.0, 1.0, 1.0],
        [2.0, -1.0, -1.0],
        [0.0, 1.0, -1.0],
        [0.0, -1.0, 1.0],
    ])

    mu1 = pv.UnstructuredGrid(cells.flatten(), celltypes, points)

    cells2 = np.array([[4, 0, 1, 2, 3],
                       [4, 4, 5, 6, 7],
                       [4, 1, 2, 3, 4]])
    celltypes2 = [pv.CellType.TETRA,
                  pv.CellType.TETRA,
                  pv.CellType.TETRA]
    mu2 = pv.UnstructuredGrid(cells2.flatten(), celltypes2, points)

    try:
        COUL.compare_mesh(mu1, mu2)
    except:
        print("Try to compare mesh unstructured with différent number of cell")

    COUL.compare_mesh(mu1, mu1)

    filed_c_1 = np.arange(mu1.number_of_cells, dtype=np.int32)
    filed_p_1 = np.arange(mu1.number_of_points, dtype=np.int32)
    mu1.cell_data["f_1"] = filed_c_1
    mu1.point_data["f_1"] = filed_p_1

    trace_c = [1, 0]
    trace_p = [4, 5, 6, 7,
               0, 1, 2, 3]

    new_pts = points[trace_p]
    new_cells = cells[trace_c]
    mu3 = pv.UnstructuredGrid(new_cells.flatten(), celltypes, new_pts)
    mu3.cell_data["f_1"] = filed_c_1[trace_c]
    mu3.point_data["f_1"] = filed_p_1[trace_p]

    COUL.compare_mesh(mu1, mu3,
                      trace_cell_m1_to_m2=trace_c,
                      trace_point_m1_to_m2=trace_p,
                      verbose=True)
    new_printer = COUL.printer()
    new_printer.mention =  "-->"
    new_printer.color = new_printer.Fore.BLUE
    new_printer.print("Coucou")

    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-05-31 13:47:33.801361"
__version__ = "@COSTO_VERSION@"
# **************************************************
import pyvista as pv
import numpy as np
import mpi4py.MPI as MPI
from scipy.spatial import KDTree
import inspect
from .mesh import Imesh
from . import utils as COUT


class Iconnector(COUT.utils):
    """
    Generique connector interface
    define the methode compute_graph that is calablel only once by construction
    All child class must overwrite the _build_graph methode
    """
    def __init__(self, univers: MPI.Comm, *args, **kwds):
        super().__init__(*args, **kwds)
        self.univers = univers
        self.neibors = []
        self._is_graph_computed = False

    # @utils.call_only_once
    def compute_graph(self, *args, **kwds):
        self._build_graph(*args, **kwds)
        self._is_graph_computed = True

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def _build_graph(self, *args, **kwds):
        raise RuntimeError("Must be overide!")


class Iconnector_using_a_mesh(Iconnector):
    """
    Connector using a mesh.
    It define its dimenions, its bounding box and its center
    """
    def __init__(self,  *args, mesh: pv.UnstructuredGrid = None,  **kwds):
        super().__init__(*args, **kwds)
        self._imesh = mesh
        self.mesh: Imesh = None
        self._init_mesh()

    def _init_mesh(self):
        self._check_input_mesh_type()

    def _check_input_mesh_type(self):

        match type(self._imesh).__name__:
            case  "UnstructuredGrid":
                self.mesh = Imesh(self._imesh)
            case  "Imesh":
                self.mesh = self._imesh
            case _:
                try:
                    self.mesh = Imesh(self._imesh)
                except:
                    raise NotImplementedError("Mesh types {} is not implemented".format(type(self._imesh)))
        return


class Iconnector_using_points(Iconnector):
    """
    Connector using points and build graphe using mathcing poitns
    """
    def __init__(self,  *args, **kwds):
        super().__init__(*args, **kwds)
        self._pts = None
        self._idx_in_local_pts = {}
        self._idx_in_neibor_pts = {}

    def _set_points(self):
        raise NotImplementedError("Must be overwrited")

    def build_base(self, graph_at_node):
        un = np.unique(graph_at_node)
        un = un[un > 0].tolist()
        res = [np.binary_repr(i) for i in un]
        lst = []
        for i in res:
            lst.append(list())
            ct = len(i)-1
            for c in i:
                if int(c) > 0:
                    lst[-1].append(ct)
                ct-=1
        base = dict(zip(un, lst))
        return base


    def _exchange_points(self):
        n_proc = self.univers.Get_size()
        rank = self.univers.Get_rank()

        pts = self._pts.flatten()

        size = np.zeros(1, dtype=np.int32)
        size[0] = pts.shape[0]
        sizes = np.empty(n_proc, dtype=np.int32)
        self.univers.Allgather(size, sizes)
        max_size = np.max(sizes)

        self.all_points = self.univers.allgather(pts)
        for i, size in enumerate(sizes):
            self.all_points[i].shape = (-1, 3)

    def _compute_intersections(self):
        m_rank = self.univers.Get_rank()
        self._set_points()
        self._exchange_points()
        kdt = KDTree(self._pts)
        # TO DO UTILISER UNE BOUNDING BOX
        for rk, interf_pts in enumerate(self.all_points):
            if rk != m_rank:
                idx_in_neibors_pts, local_trace = self._compute_pts_intersection(kdt, interf_pts)

                # a sort is needed to identically  order points in each side
                to_sort = None
                if rk < m_rank:
                    to_sort = idx_in_neibors_pts
                else:
                    to_sort = local_trace
                trace_common_order = np.argsort(to_sort)

                if local_trace.shape[0] > 0:
                    self.neibors.append(rk)
                    self._idx_in_neibor_pts[rk] = idx_in_neibors_pts[trace_common_order]
                    self._idx_in_local_pts[rk] = local_trace[trace_common_order]

    def _compute_pts_intersection(self, kdt, pts):
        dist, local_trace = kdt.query(pts, eps=1e-7, distance_upper_bound=1e-3)
        idx_in_pts = np.argwhere(local_trace < kdt.n).flatten()
        local_trace = local_trace[idx_in_pts].flatten()
        return idx_in_pts, local_trace

    def _build_graph(self, *args, **kwds):
        self._compute_intersections()
        self._add_graph_fields(*args, **kwds)

    def _add_graph_fields(self, *args, **kwds):
        raise NotImplementedError("Must be overwrited")




class connect_skin_centers(Iconnector_using_points, Iconnector_using_a_mesh):
    """
    build the connection between meshes  based on the skin centers of each meshs
    To establish it, the interface of the mesh is build and its cell center are use to
    find matching element
    """
    def __init__(self,  *args, **kwds):
        super().__init__(*args, **kwds)
        self._pts = None
        self.interface = None
        self.trace_cell_centers = {}

    def _set_points(self):
        self.mesh.compute_interface_cell_centers()
        self.interface = self.mesh.interface
        self._pts = self.mesh.interface_pts

    def _add_graph_fields(self, *args, fields_racine="graph", add_individual_graphs=False,
                          build_nodal_graph=False, **kwds):

        # check that the new fields name (@cell and @ point) is not already exists
        c_keys = list(self.interface.cell_data.keys())
        p_keys = list(self.interface.point_data.keys())

        if (fields_racine in c_keys) or (fields_racine in p_keys):
            raise RuntimeError("{} already exists in the interface mesh\n\
            Known fileds are:\n\
            @cell:\n{}\n\
            @node:\n{}\n".format(fields_racine, c_keys, p_keys))

        # add a temporary field @ nodes for index
        tmp_index_name = "tmp_index"
        while tmp_index_name in p_keys:
            tmp_index_name += "_tmp"
        tmp_idx = np.arange(self.interface.number_of_points, dtype=np.int32)
        self.interface.point_data[tmp_index_name] = tmp_idx

        if build_nodal_graph:
            add_individual_graphs = True
        # add the individual and global interface connexion field @cell
        glob_connection = np.zeros(self.interface.number_of_cells, dtype=np.int32)
        for rk, idx in self._idx_in_local_pts.items():
            glob_connection[idx] += 2**rk
            if add_individual_graphs:
                connection = np.zeros(self.interface.number_of_cells, dtype=np.int32)
                connection[idx] = 1
                field_name = fields_racine+"_{}".format(rk)
                self.interface.cell_data[field_name] = connection
            self.trace_cell_centers[rk] = self.interface.cell_data[self.mesh._field_ids_name][idx]
        self.interface.cell_data[fields_racine] = glob_connection

        if build_nodal_graph:
            self._build_nodal_graph(tmp_index_name, fields_racine)

        del self.interface.point_data[tmp_index_name]
        del self.interface.cell_data[self.mesh._field_ids_name]
        del self.interface.point_data[self.mesh._field_ids_name]

        del self.mesh.cell_data[self.mesh._field_ids_name]
        del self.mesh.point_data[self.mesh._field_ids_name]

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def _build_nodal_graph(self, tmp_index_name="tmp_index", fields_racine="graph"):
        # add the global connexion field @point
        connexion_at_pts = np.zeros(self.interface.number_of_points, dtype=np.int32)
        eps = 1e-7
        for rk in self._idx_in_neibor_pts.keys():
            c_field_name = fields_racine+"_{}".format(rk)
            interf = self.interface.threshold(value=[1-eps, 1+eps], scalars=c_field_name)
            idx = interf.point_data[tmp_index_name]
            connexion_at_pts[idx] += 2**rk
        self.interface.point_data[fields_racine] = connexion_at_pts

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing


class connect_skin_points(Iconnector_using_points, Iconnector_using_a_mesh):
    """
    build the connection between meshes  based on the skin centers of each meshs
    To establish it, the interface of the mesh is build and its cell center are use to
    find matching element
    """
    def __init__(self,  *args, **kwds):
        super().__init__(*args, **kwds)
        self._pts = None
        self.interface = None
        self.traces_in_local_interface = {}
        self.traces_in_neibor_interfaces = {}
        self.traces_in_mesh = {}
        self.map_mesh_to_interface_node = None
        self.map_interface_to_mesh_node = None
        self._fields_racine = None
        self._ret = None

    def _set_points(self):
        self.mesh.build_skin()
        self.interface = self.mesh.interface
        self._pts = self.interface.points

    def _add_graph_fields(self, *args, fields_racine="graph", add_individual_graphs=False, **kwds):
        # check that the new fields name (@cell and @ point) is not already exists
        c_keys = list(self.interface.cell_data.keys())
        p_keys = list(self.interface.point_data.keys())
        # for rk in self._pts_idx.keys():
        #     field_name = fields_racine+"_{}".format(rk)
        #     if (field_name in c_keys) or (field_name in p_keys):
        #         raise RuntimeError("{} already exists in the interface mesh\n\
        #         Known fileds are:\n\
        #         @cell:\n{}\n\
        #         @node:\n{}\n".format(field_name, c_keys,p_keys))
        self._fields_racine = fields_racine

        if (fields_racine in c_keys) or (fields_racine in p_keys):
            raise RuntimeError("{} already exists in the interface mesh\n\
            Known fileds are:\n\
            @cell:\n{}\n\
            @node:\n{}\n".format(fields_racine, c_keys, p_keys))
        # add a temporary field @ nodes for index
        tmp_index_name = "tmp_index"
        while tmp_index_name in p_keys:
            tmp_index_name += "_tmp"
        tmp_idx = np.arange(self.interface.number_of_points, dtype=np.int32)
        self.interface.point_data[tmp_index_name] = tmp_idx

        mesh_ids_in_interface = self.interface.point_data[self.mesh._field_ids_name]
        # res = pv.convert_array(mesh_ids_in_interface)
        # print(res)
        # print(mesh_ids_in_interface)

        self.map_mesh_to_interface_node = dict(zip(mesh_ids_in_interface, tmp_idx))
        self.map_interface_to_mesh_node = mesh_ids_in_interface

        # # add the individual and global interface connexion field @cell
        # glob_connection = np.zeros(self.interface.number_of_cells, dtype=np.int32)
        # for rk, idx in self._pts_idx.items():
        #     glob_connection[idx] += 2**rk
        #     connection = np.zeros(self.interface.number_of_cells, dtype=np.int32)
        #     connection[idx] = 1
        #     field_name = fields_racine+"_{}".format(rk)
        #     self.interface.cell_data[field_name] = connection
        #     self.trace_cell_centers[rk] = self.interface.cell_data[self.mesh._field_ids_name][idx]

        # self.interface.cell_data[fields_racine] = glob_connection

        # add the global connexion field @point
        connexion_at_pts = np.zeros(self.interface.number_of_points, dtype=np.int32)
        eps = 1e-7
        for rk, idx_in_local_pts in self._idx_in_local_pts.items():
            if add_individual_graphs:
                connection = np.zeros(self.interface.number_of_points, dtype=np.int32)
                connection[idx_in_local_pts] = 1
                field_name = fields_racine+"_{}".format(rk)
                self.interface.point_data[field_name] = connection
            connexion_at_pts[idx_in_local_pts] += 2**rk
            self.traces_in_local_interface[rk] = idx_in_local_pts
            self.traces_in_mesh[rk] = self.interface.point_data[self.mesh._field_ids_name][idx_in_local_pts]
            self.traces_in_neibor_interfaces[rk] = self._idx_in_neibor_pts[rk]

        self.interface.point_data[fields_racine] = connexion_at_pts
        del self.interface.point_data[tmp_index_name]
        if tmp_index_name in self.interface.cell_data.keys():
            del self.interface.cell_data[self.mesh._field_ids_name]
        del self.interface.point_data[self.mesh._field_ids_name]

        if self.mesh._field_ids_name in self.mesh.cell_data.keys():
            del self.mesh.cell_data[self.mesh._field_ids_name]
        del self.mesh.point_data[self.mesh._field_ids_name]


        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def _project_connectivity_graph_by_neibors(self):
        if self._is_graph_computed is False:
            raise RuntimeError("Graph must be computed first")
        self._ret = {}
        graph_at_node = self.interface.point_data[self._fields_racine]
        self._ret["base"] = self.build_base(graph_at_node)
        # for i, trace in self.traces_in_local_interface.items():
        #     print("trace_point in local interface (neibor rank):", i, trace)

        # for i, trace in self.traces_in_neibor_interfaces.items():
        #     print("trace_point in neibor interface (neibor rank):", i, trace)

        # for i, trace in self.traces_in_mesh.items():
        #     print("trace_point in mesh id rank:", i, trace)
        #     print("trace_point in global id rank:", i, self.mesh.point_data["CHECK_GLOBAL_IDS"][trace])
        # print("base:", base)
        # print("graph:\n",graph_at_node)

        trace_in_interface_mesh = {}
        trace_in_neibor_interface = {}
        trace_in_mesh = {}
        for compo, neibors in self._ret["base"].items():
            to_take = np.array(graph_at_node == compo, dtype=bool)
            candidat = None
            if len(neibors) == 1:
                candidat = neibors[0]
            else:
                candidat = min(neibors)
            trace = self.traces_in_local_interface[candidat]
            filtre = to_take[trace]

            trace_in_interface_mesh[compo] = trace[filtre]

            trace = self.traces_in_neibor_interfaces[candidat]
            trace_in_neibor_interface[compo] = trace[filtre]

            trace = self.traces_in_mesh[candidat]
            trace_in_mesh[compo] = trace[filtre]
        self._ret["trace_in_interface_mesh"] = trace_in_interface_mesh
        self._ret["trace_in_neibor_interface"] = trace_in_neibor_interface
        self._ret["trace_in_mesh"] = trace_in_mesh

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def get_traces_and_map(self):
        if self._ret is None:
            self. _project_connectivity_graph_by_neibors()
        return self._ret


class connect_points(Iconnector_using_points):
    """
    build the connection between meshes  based on the skin centers of each meshs
    To establish it, the interface of the mesh is build and its cell center are use to
    find matching element
    """
    def __init__(self,  mesh, *args, **kwds):
        super().__init__(*args, **kwds)
        self.interface = mesh
        self.traces_in_local_interface = {}
        self.traces_in_neibor_interfaces = {}
        self._is_graph_computed = False
        self.traces_in_mesh = {}
        self._ret = None

    def _set_points(self):
        assert(type(self.interface) == pv.PolyData)
        self._pts = self.interface.points


    def _add_graph_fields(self, *args, fields_racine="graph", add_individual_graphs=False, **kwds):
        self.idx_in_neibor_pts = self._idx_in_neibor_pts
        self.idx_in_local_pts = self._idx_in_local_pts


        c_keys = list(self.interface.cell_data.keys())
        p_keys = list(self.interface.point_data.keys())
        # for rk in self._pts_idx.keys():
        #     field_name = fields_racine+"_{}".format(rk)
        #     if (field_name in c_keys) or (field_name in p_keys):
        #         raise RuntimeError("{} already exists in the interface mesh\n\
        #         Known fileds are:\n\
        #         @cell:\n{}\n\
        #         @node:\n{}\n".format(field_name, c_keys,p_keys))
        self._fields_racine = fields_racine

        if (fields_racine in c_keys) or (fields_racine in p_keys):
            raise RuntimeError("{} already exists in the interface mesh\n\
            Known fileds are:\n\
            @cell:\n{}\n\
            @node:\n{}\n".format(fields_racine, c_keys, p_keys))
        # add a temporary field @ nodes for index
        tmp_index_name = "tmp_index"
        while tmp_index_name in p_keys:
            tmp_index_name += "_tmp"
        tmp_idx = np.arange(self.interface.number_of_points, dtype=np.int32)
        self.interface.point_data[tmp_index_name] = tmp_idx

        connexion_at_pts = np.zeros(self.interface.number_of_points, dtype=np.int32)
        eps = 1e-7
        for rk, idx_in_local_pts in self._idx_in_local_pts.items():
            # if add_individual_graphs:
            #     connection = np.zeros(self.interface.number_of_points, dtype=np.int32)
            #     connection[idx_in_local_pts] = 1
            #     field_name = fields_racine+"_{}".format(rk)
            #     self.interface.point_data[field_name] = connection
            connexion_at_pts[idx_in_local_pts] += 2**rk
            self.traces_in_local_interface[rk] = idx_in_local_pts
            self.traces_in_neibor_interfaces[rk] = self._idx_in_neibor_pts[rk]

        self.interface.point_data[fields_racine] = connexion_at_pts
        del self.interface.point_data[tmp_index_name]

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing


    def _project_connectivity_graph_by_neibors(self):
        if self._is_graph_computed is False:
            raise RuntimeError("Graph must be computed first")
        self._ret = {}
        graph_at_node = self.interface.point_data[self._fields_racine]
        self._ret["base"] = self.build_base(graph_at_node)
        trace_in_neibor_interface = {}
        trace_in_mesh = {}
        for compo, neibors in self._ret["base"].items():
            to_take = np.array(graph_at_node == compo, dtype=bool)
            candidat = None
            if len(neibors) == 1:
                candidat = neibors[0]
            else:
                candidat = min(neibors)
            trace = self.traces_in_local_interface[candidat]
            filtre = to_take[trace]
            trace_in_mesh[compo] = trace[filtre]

            trace = self.traces_in_neibor_interfaces[candidat]
            # trace_in_neibor_interface[compo] = trace[filtre]


            # trace_in_mesh[compo] = np.argwhere(graph_at_node == compo).flatten()

        # self._ret["trace_in_interface_mesh"] = trace_in_interface_mesh
        # self._ret["trace_in_neibor_interface"] = trace_in_neibor_interface
        self._ret["trace_in_mesh"] = trace_in_mesh

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def get_traces_and_map(self):
        if self._ret is None:
            self. _project_connectivity_graph_by_neibors()
        return self._ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-11-09 17:39:32.284214"
__version__ = 1.0

import costo.scheduler as COS
from costo.clock import clock as Clock
import traceback

def sould_not_be_called():
    raise RuntimeError

def test_IScheduler():
    clock = Clock(0.)

    clock.reset(0.2)

    scheduler = COS.IScheduler()
    scheduler.planify_new_meeting(1.)
    scheduler.accept()
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False

    clock.tic(0.8)
    assert clock.time == 0.2
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False
    clock.tac()
    assert clock.time == 1.
    assert scheduler.is_a_meeting_point  == True
    assert scheduler.do_publish_action()  == True
    assert scheduler.do_collect_action()  == True


    scheduler.planify_new_meeting(2.)
    scheduler.accept()
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False

    clock.tic(1.)
    assert clock.time == 1.
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False
    clock.tac()
    assert clock.time == 2.
    assert scheduler.is_a_meeting_point  == True
    assert scheduler.do_publish_action()  == True
    assert scheduler.do_collect_action()  == True

    clock.reset(0.)
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False

    clock.tic(2.)
    assert clock.time == 0.
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False
    clock.tac()
    assert clock.time == 2.
    assert scheduler.is_a_meeting_point  == True
    assert scheduler.do_publish_action()  == True
    assert scheduler.do_collect_action()  == True

    clock.tic(2.)
    # reset between tic and tac calls
    clock.reset(0.)

    try:
        clock.tac()
        sould_not_be_called()
    except:
        print("Normal exception (tic must be call first)")
    # A tic call is needed first
    clock.tic(2.)
    assert clock.time == 0.
    assert scheduler.is_a_meeting_point  == False
    assert scheduler.do_publish_action()  == False
    assert scheduler.do_collect_action()  == False
    clock.tac()
    assert clock.time == 2.
    assert scheduler.is_a_meeting_point  == True
    assert scheduler.do_publish_action()  == True
    assert scheduler.do_collect_action()  == True

    clock.tic(2.)
    try:
        clock.tic(1.)
        sould_not_be_called()
    except:
        print("Normal exception (tic can't be called two times without a tac call between)")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-07-03 18:09:05.044015"
__version__ = "@COSTO_VERSION@"
# **************************************************
import mpi4py.MPI as MPI
import numpy as np
import inspect
from .scheduler import IScheduler

class IInternalNeibor():
    def __init__(self,
                 comm: MPI.Comm = MPI.COMM_NULL,
                 rank: int = 0,
                 # size: int = 0,
                 # dim: int = 0,
                 buff_send: np.array = None,
                 buff_recv: np.array = None,
                 ):
        self._comm = comm
        self.rank = rank
        self.Ar = None
        self.Br = None
        self._tag_send_Ar_Br = 200
        self._tag_recv_Ar_Br = 200
        self._buff_send = buff_send
        self._buff_recv = buff_recv
        self._send_reqs = 0
        self._recv_req = 0

    def _send_Ar_Br(self):
        """
        common send methode
        """
        req = self._comm.Isend(self._buff_send,
                               dest=self.rank,
                               tag=self._tag_send_Ar_Br)
        return req

    def _recv_Ar_Br(self):
        """
        common recv methode
        """
        self._recv_req = self._comm.Recv(self._buff_recv,
                                         source=self.rank,
                                         tag=self._tag_recv_Ar_Br)
        return

    def publish(self):
        self._send_reqs = self._send_Ar_Br()

    def collect(self):
        self._recv_Ar_Br()


class INeibor(IInternalNeibor):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.scheduler:IScheduler = None

    def publish(self,):
        raise NotImplementedError("You must set a scheduler first")

    def collect(self,):
        raise NotImplementedError("You must set a scheduler first")

    def set_scheduler(self, scheduler:IScheduler=None):
        self.scheduler = scheduler
        self.publish = self._publish_ok
        self.collect = self._collect_ok


    def _publish_ok(self):
        if self.scheduler.is_a_meeting_point:
            self._send_reqs = self._send_Ar_Br()

    def _collect_ok(self):
        if self.scheduler.do_collect_action():
            self._recv_Ar_Br()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-07-03 17:28:00.999352"
__version__ = "@COSTO_VERSION@"
# **************************************************
from . import utils as COUT


class clock(metaclass=COUT.SingletonMeta):
    __slots__ = (
        "_time",
        "_dt",
        "_tic_call_allowed",
        "_tac_call_allowed"
    )

    def __init__(self, *args, t0=0.,**kwds):
        self._time = float(t0)
        self._dt = 0.
        self._tic_call_allowed = True
        self._tac_call_allowed = False

    @property
    def time(self):
        """
        return the time
        """
        return self._time

    @time.setter
    def time(self, val):
        raise RuntimeError("Time can't be set, use reset or increment")

    def reset(self, t):
        """
        reset the clock to time t
        """
        self._time = float(t)
        self._tic_call_allowed = True
        self._tac_call_allowed = False
        return

    def tic(self, dt):
        """
        set the elapsed time (dt) until the call of tac method
        """
        if self._tic_call_allowed:
            self._dt = dt
            self._tac_call_allowed = True
            self._tic_call_allowed = False
            return
        else:
            raise RuntimeError("You must call tac method before this call.")

    def tac(self):
        """
        Advance the time with the dt set on tic methode
        """
        if self._tac_call_allowed:
            self._time += self._dt
            self._dt = 0.
            self._tic_call_allowed = True
            self._tac_call_allowed = False
            return
        else:
            raise RuntimeError("You must call tac method before")

    @property
    def next_time(self):
        """
        If this method is called between a tic and tac calls
        its return the next time of the clock.
        It meens the time that will be when tac will be called.
        If this method is called between a tac and tic calls
        the current time is return. We don't know yet the next time step.
        """
        if self._tac_call_allowed:
            # between tic and tac
            return self._time + self._dt
        else:
            # between tac and tic
            return self._time

    @next_time.setter
    def next_time(self, val):
        raise RuntimeError("next_time can't be set")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-07-03 17:28:00.999352"
__version__ = "@COSTO_VERSION@"
# **************************************************
from .clock import clock

class IScheduler():
    __slots__ = (
        "_clock",
        "_its_a_meeting_point",
        "_dt_to_meeting",
        "_dt_seuil",
        "_t_next_meeting",
        "_new_t_meeting",
        "_planify_call_allowed",
        "_accept_call_allowed"
    )

    def __init__(self, dt_seuil=1e-9):
        self._clock = clock()
        self._its_a_meeting_point = True
        self._dt_to_meeting = -1
        self._dt_seuil = dt_seuil
        # self.neibor_dt = 0
        # self._dtmax = 0
        # self._dtmin = 0
        self._t_next_meeting = 0
        self._new_t_meeting = 0

        self._planify_call_allowed = True
        self._accept_call_allowed = False
        # self.dt_guest = 0

    @property
    def is_a_meeting_point(self):
        self._dt_to_meeting = abs(self.t_next_meeting-self._clock.time)
        if self._dt_to_meeting < self._dt_seuil:
            self._its_a_meeting_point = True
        else:
            self._its_a_meeting_point = False
        return self._its_a_meeting_point

    @is_a_meeting_point.setter
    def is_a_meeting_point(self, val):
        return

    def planify_new_meeting(self, val):
        if self._planify_call_allowed:
            self._new_t_meeting = val
            self._planify_call_allowed = False
            self._accept_call_allowed = True
        else:
            raise RuntimeError("A new meeting has been planified and not yet accepted")
        return

    def accept(self):
        if self._accept_call_allowed:
            self._t_next_meeting = self._new_t_meeting*1.
            self._accept_call_allowed = False
            self._planify_call_allowed = True
        else:
            raise RuntimeError("You must call planify_new_meeting before")

    @property
    def t_next_meeting(self):
        return self._t_next_meeting

    @t_next_meeting.setter
    def t_next_meeting(self, val):
        raise RuntimeError("Yout must use planify_new_meeting and accept to modify it")

    def do_publish_action(self):
        return self.is_a_meeting_point

    def do_collect_action(self):
        # self._check_schedule()  # par securité le garder?
        return self._its_a_meeting_point



    def __repr__(self):
        to_ret = ""
        to_ret += "time to next meeting: " + str(self._dt_to_meeting) + "\n"
        to_ret += "next meeting point: " + str(self.t_next_meeting) + "\n"
        return to_ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-06-26 08:16:48.772672"
__version__ = "@COSTO_VERSION@"
# **************************************************

import numpy as np
import inspect
from . import utils as COUT
from .clock import clock
import mpi4py.MPI as MPI


class ItimeStepStrategy():
    def __init__(self,  *args, **kwds):
        self.name = self.__class__.__name__
        self.cpl = None

    def set_cpl(self,cpl):
        self.cpl = cpl
        print("cpl:", self.cpl.c_data)

    def get_next_dt(self, *args, **kwds):
        raise NotImplementedError("Must be override")
        # self._max_dt_local = self.c_data.comm_l.allreduce(dt, op=MPI.MIN)

    def __repr__(self):
        to_ret = "Strategy class used is: "+ str(self.name)
        to_ret +="\n "
        return to_ret

class useTheSmallestConstrain(ItimeStepStrategy):
    def __init__(self, *args, **kwds):
        super().__init__(self)
        self._dt = 0

    def get_next_dt(self,  *args, max_dt_local=0, **kwds):
        self._dt = self.cpl.c_data.comm_w.allreduce(max_dt_local, op=MPI.MIN)
        return self._dt


class CoupledTimeStepComputer(ItimeStepStrategy):
    def __init__(self, *args, **kwds):
        ItimeStepStrategy.__init__(self, *args, **kwds)
        self._clock = clock()
        self.next_meeting_point = 0
        self.dt_eps = 1e-9
        self.dt_local_max = 0
        self._delta_close = 0.5
        self.all_dt_max = []
        self.all_other_dt_max = []
        self._the_fist_meeting = True
        self.dt = 0

    def get_next_dt(self, max_dt_local):
        self.dt = self._compute_dt(max_dt_local)
        return self.dt, self.next_meeting_point

    def _compute_dt(self, max_dt_local):
        self.dt_local_max = self.cpl.c_data.comm_l.allreduce(max_dt_local, op=MPI.MIN)
        if (self._the_fist_meeting):
            self._the_fist_meeting = False
            return self._determine_next_meeting_point()

        dt_to_next_meeting_point = abs(self.next_meeting_point - self._clock.time)

        if dt_to_next_meeting_point < self.dt_eps:
            # there is a meeting points
            return self._determine_next_meeting_point()
        else:
            # continue until next meeting points

            if self.dt_local_max < dt_to_next_meeting_point:
                # advance with the largest available time step
                return self.dt_local_max
            else:
                # adapt last time step to reach next meeting point
                return dt_to_next_meeting_point


    def _determine_next_meeting_point(self):
        # communication globale possible au début
        self.all_dt_max = self.cpl.c_data.comm_w.allgather(self.dt_local_max)
        self.all_dt_max = np.array(self.all_dt_max)
        self.all_other_dt_max = np.unique(self.all_dt_max[self.cpl.other_code_ranks])
        assert len(self.all_other_dt_max) == 1
        all_time_step_ratio = self.all_dt_max/self.dt_local_max
        trace_less_one_ratio = np.argwhere(all_time_step_ratio <= 1)
        unify_time_step = all_time_step_ratio.copy()
        unify_time_step[trace_less_one_ratio] = 1/all_time_step_ratio[trace_less_one_ratio]

        ratio = self.all_other_dt_max[0]/self.dt_local_max
        if ratio <= 1:
            ratio = 1/ratio
        # if verbose > 0:
        #     print("time step ratio: ", ratio, ", time: ", self._clock.time)
        rank_close = []
        if (abs(ratio-1) < self._delta_close):
            rank_close = [-1]
        dt = 0
        if len(rank_close) > 0:
            # use same time step
            dt = np.min(self.all_dt_max)
            self.next_meeting_point = dt*1.0 + self._clock.time
        else:
            max_dt = np.max(self.all_dt_max)
            min_dt = np.min(self.all_dt_max)
            if abs(self.dt_local_max - max_dt)/max_dt < self.dt_eps:
                dt = max_dt
            else:
                dt = min_dt
            self.next_meeting_point = max_dt + self._clock.time
        return dt


class TimeStepComputer(COUT.utils):
    def __init__(self, cpl, *args, **kwds):
        self.cpl = cpl
        self.timeStepStrategy: ItimeStepStrategy = None
        self._is_time_step_strategie_set = False

    def get_next_dt(self, *args, **kwds):
        raise RuntimeError("You must set a strategy to compute time step")

    def set_strategy(self, *args, timeStepStrategy:ItimeStepStrategy = None, **kwds):
        subclasses = COUT.FindAllSubclasses(ItimeStepStrategy)
        subclasseNames = [subclasse[1] for subclasse in subclasses]
        if not (timeStepStrategy.__name__ in subclasseNames):
            raise NotImplementedError("Given strategy (class {}) is not available.\nAvailable candidat are:\n--> {}".format(timeStepStrategy.__name__ , "\n--> ".join(subclasseNames)))


        strategies = self.cpl.c_data.comm_w.allgather(timeStepStrategy.__name__)
        uniqueStrategy = np.unique(np.array(strategies))
        if len(uniqueStrategy) > 1:
            raise NotImplementedError("All code must use the same strategy to compute time step\n Current strategies ares {}".format(uniqueStrategy))
        self.timeStepStrategy = timeStepStrategy()
        self.timeStepStrategy.set_cpl(self.cpl)
        self.__dict__["get_next_dt"] = self.timeStepStrategy.get_next_dt
        self._is_time_step_strategie_set = True

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-11-08 08:19:49.161191"
__version__ = 1.0

import costo.mesh as COM
import pyvista as pv
import numpy as np
import os
src_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))

def read(fn):
    return pv.read(os.path.join(src_dir, fn))

def test_build_map_eltype_dim():
    omega = read("00-DATA/first_order_elements.vtu")
    dims = COM.build_map_eltype_dim(mesh=omega,
                                    verbose=True)
    REF = {'LINE': {'vtktype': 3, 'dim': 1},
           'TRIANGLE': {'vtktype': 5, 'dim': 2},
           'QUAD': {'vtktype': 9, 'dim': 2},
           'TETRA': {'vtktype': 10, 'dim': 3},
           'HEXAHEDRON': {'vtktype': 12, 'dim': 3},
           'WEDGE': {'vtktype': 13, 'dim': 3}
           }
    assert dims == REF
    return


def test_build_bulk_and_interface_elt_fields():
    omega = read("00-DATA/bulk_int_msh.vtu")
    COM.build_bulk_and_interface_elt_fields(omega=omega,
                                            bulk_elt_name="bulk_elt",
                                            interface_elt_name="interface_elt",
                                            )

    omega.save("omega.vtu")
    ref = read("00-DATA/omega.vtu")
    assert np.allclose(omega.cell_data["interface_elt"],
                       ref.cell_data["interface_elt"])
    assert np.allclose(omega.cell_data["bulk_elt"],
                       ref.cell_data["bulk_elt"])

    inter_elts = omega.threshold(scalars="interface_elt", value=[0.9, 1.1])
    bulk_elts = omega.threshold(scalars="bulk_elt", value=[0.9, 1.1])
    inter_elts.save("interface.vtu")
    bulk_elts.save("tetra.vtu")
    return


def test_build_graph_partitionning():
    omega = read("00-DATA/tetra.vtu")
    COM.build_graph_partitionning(omega=omega,
                                  nb_part=2,
                                  part_name="epart",
                                  part_name_at_node="npart")
    omega.save("split.vtu")
    ref = read("00-DATA/split.vtu")
    assert np.allclose(omega.cell_data["epart"],
                       ref.cell_data["epart"])
    assert np.allclose(omega.point_data["npart"],
                       ref.point_data["npart"])


def test_extract_subdomains():
    omega = read("00-DATA/split.vtu")
    omegas = COM.extract_subdomains(omega=omega,
                                    part_name="epart")
    omegas[0].save("domain_1.vtu")
    omegas[1].save("domain_2.vtu")
    ref_dom1 = read("00-DATA/domain_1.vtu")
    ref_dom2 = read("00-DATA/domain_2.vtu")
    assert np.allclose(ref_dom1.points,
                       omegas[0].points)

    assert np.allclose(ref_dom2.points,
                       omegas[1].points)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-05-31 09:39:36.902967"
__version__ = "@COSTO_VERSION@"
# **************************************************
from colorama import init as colorama_init
from colorama import Fore
from colorama import Style
import numpy as np
import pyvista as pv

colorama_init()


class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

def my_fct_decorer(fct):
    def new_fct(*args, **kwds):
        print("(inspect) -> call function", fct.__name__)
        return fct(*args, **kwds)
    return new_fct


def fct_tracer(cls):

    att_idct = cls.__dict__.items()
    fct_to_decorate = {}
    # print(list(cls.__dict__["__dict__"].items()))
    for name,val in att_idct:
        if not name.startswith("__"):
            fct_to_decorate[name] = my_fct_decorer(val)
            setattr(cls, name, my_fct_decorer(val))
    print(fct_to_decorate)
    return cls

def check_inputs(c_data, used_keys):
    """
    Check that keys exist in a dictionnray.
    A ValueError is raise if its not the case.

    Inputs:
       - c_data (dict) dictionnary in which to make the check
       - used_kyes (list) list of (potential) keys in c_data
    """
    assert type(c_data) == dict
    used_keys_ok = used_keys

    if type(used_keys) != list:
        used_keys_ok = [used_keys]

    input_keys = c_data.keys()
    for k in used_keys_ok:
        if not k in input_keys:
            raise ValueError("keys: {} is required, but not provied".format(k))
    return


class printer():
    """
    Usefull class to make beautifull print.

    Members ares:
       - mention: some additional mention in the begining of the print
       - color: the color
       - Fore: the containre of color (from colorama module)
       - print: the new print methode to use.

    Example:
    new_printer = printer()
    new_printer.mention =  "-->"
    new_printer.color = new_printer.Fore.BLUE
    new_printer.print("Coucou")
    """
    def __init__(self, mention: str =""):
        self.old_print = print
        self.mention = mention
        self.Fore = Fore
        self.color = Fore.BLUE
        self.print = self._new_print

    def _new_print(self, *args, **kwargs):
        new_args = [self.color+str(self.mention)]
        for arg in args:
            new_args.append(arg)
        new_args.append(""+Fore.RESET)
        return self.old_print(*new_args, **kwargs)


class utils():
    """
    Usefull class to drevive from that contain a do_nothing method.
    """

    def call_only_once(fct):
        """
        Use at decorer to make a function callable only once.
        """
        def new_fct(*args, **kwds):
            print("Call function", fct.__name__)
            fct(*args, **kwds)
            return
        return new_fct
        # # mecanisme to call only once this fonction
        # self_fct_name = inspect.currentframe().f_code.co_name
        # self.__dict__[self_fct_name] = self.do_nothing

    def __init__(self):
        return

    def do_nothing(self, *args, **kwds):
        """
        Methode that does nothing
        """
        return



def map_data_on_local_BC(kdt, interface, fields_to_transfert: list=[None]):
    pts = interface.cell_centers().points

    dd, ii = kdt.query(pts, eps=1e-3)
    # print(kdt.n)
    # print(dd)
    # print(pts.shape[0])
    # print(ii)
    # trace = np.argwhere( ii != kdt.n).flatten()
    trace = np.argwhere(dd < 1e-7).flatten()
    ii = ii[trace]
    for ifiled in fields_to_transfert["cell"]:
        test = np.ones(interface.number_of_cells, dtype=np.int32) * -1
        test[trace] = ifiled[1][ii]
        interface.cell_data[ifiled[0]] = test


def FindAllSubclasses(classType):
    """
    Build a list of tuple wich contain className and Class instance for each child Class of classType.

    Input:
       - classType: The mother class whose child are looked.
    """
    import sys
    import inspect
    subclasses = []
    callers_module = sys._getframe(1).f_globals['__name__']
    classes = inspect.getmembers(sys.modules[callers_module], inspect.isclass)
    for name, obj in classes:
        if (obj is not classType) and (classType in inspect.getmro(obj)):
            subclasses.append((obj, name))
    return subclasses


def compare_mesh(m1, m2, trace_cell_m1_to_m2=None, trace_point_m1_to_m2=None, use_fields_from_m1_only=False, verbose=0):
    """
    compare two pyvista meshs (if theirs differt, an assert is raised).

    Availabale mesh types are:
       - pv.core.pointset.UnstructuredGrid
       - pv.core.pointset.StructuredGrid
       - pv.core.pointset.PolyData

    can you cast it ?

    Inputs:
       - m1: mesh 1
       - m2: mesh 2
       - trace_cell_m1_to_m2: optional trace to map cell_data from m1 to m2
       - trace_point_m1_to_m2: optional trace to map poitn_data from m1 to m2
       - use_fields_from_m1_only (bool): active a mode that only check if fields from m1 exits in m2
       - verbose (bool): active / desactive verbosity
    """
    import pyvista as pv
    trace_c_set = False
    trace_p_set = False


    # --------------------------------------------------
    # global checks
    # --------------------------------------------------
    # same mesh type ?
    if not type(m1) == type(m2):
        raise RuntimeError("Mesh type differs.\n \
        \t->type mesh1: {}\n\
        \t->type mesh2: {}".format(type(m1),
                                   type(m2)))

    # type ok?
    if type(m1) not in [pv.core.pointset.UnstructuredGrid,
                        pv.core.pointset.StructuredGrid,
                        pv.core.pointset.PolyData]:
        raise NotImplementedError("Type of m1 is not available ({})\n".format(type(m1))+
        "You can try to cast them into one available type?\n"+
        "see DataSet.cast_to_unstructured_grid()\n"+
        "or DataSet.cast_to_poly_points()\n"+
        "or DataSet.cast_to_pointset() \n"+
        "https://docs.pyvista.org/api/core/_autosummary/pyvista.dataset")

    # same number of point ?
    if not m1.number_of_points == m2.number_of_points:
        raise RuntimeError("Number of point differs.\n \
        \t-> m1: {}\n\
        \t-> m2: {}".format(m1.number_of_points,
                            m2.number_of_points))

    if trace_point_m1_to_m2 is None:
        trace_p_set = True
        trace_point_m1_to_m2 = np.arange(m1.number_of_points, dtype=np.int32)

    # same number of point fields ?
    if not use_fields_from_m1_only:
        if not len(m1.point_data) == len(m2.point_data):
            raise RuntimeError("Number of point data are different\n\
            \t-> m1: {}\n\
            \t-> m2: {}".format(len(m1.point_data),
                                len(m2.point_data)
                                ))

    # same points ?
    if not np.allclose(m1.points[trace_point_m1_to_m2], m2.points):
        raise RuntimeError("Point of the meshs differs")

    # same number of cells ?
    if not m1.number_of_cells == m2.number_of_cells:
        raise RuntimeError("Number of cell are differents\n\
        \t-> m1: {}\n\
        \t-> m2: {}".format(m1.number_of_cells,
                            m2.number_of_cells
                            ))

    # same number of cells data ?
    if not use_fields_from_m1_only:
        if not len(m1.cell_data) == len(m2.cell_data):
            raise RuntimeError("Number of cell data are differents\n\
            \t-> m1: {}\n\
            \t-> m2: {}".format(len(m1.cell_data),
                                len(m2.cell_data)
                                ))


    # --------------------------------------------------

    # --------------------------------------------------
    # specifique checks (mesh type)
    # --------------------- -----------------------------
    if type(m1) is pv.core.pointset.UnstructuredGrid:
        if trace_cell_m1_to_m2 is None:
            trace_c_set = True
            trace_cell_m1_to_m2 = np.arange(m1.number_of_cells, dtype=np.int32)
        if not use_fields_from_m1_only:
            if trace_c_set is True:
                assert np.allclose(m1.cells, m2.cells)
    if type(m1) is pv.core.pointset.PolyData:
        if trace_cell_m1_to_m2 is None:
            trace_cell_m1_to_m2 = np.arange(m1.n_cells, dtype=np.int32)
        assert len(m1.faces) == len(m2.faces)
        if not use_fields_from_m1_only:
            assert np.allclose(m1.faces, m2.faces)

    # --------------------------------------------------
    # checks fields
    # --------------------------------------------------
    # @ cells
    for k in m1.cell_data:
        k1 = k
        if verbose > 0:
            print("compare {} in cell_data".format(k))
        # if k == "rhoc" or k == "F" or k == "W" or k == "M_inv":
        #     continue
        # if k == "E":
        #     k1 = "total-energy"
        if not np.allclose(m1.cell_data[k][trace_cell_m1_to_m2], m2.cell_data[k1]):
            print("Diff in {} in cell_data".format(k))
            if verbose > 0:
                print(m1.cell_data[k][trace_cell_m1_to_m2]- m2.cell_data[k1])
                print(np.linalg.norm(m1.cell_data[k][trace_cell_m1_to_m2]- m2.cell_data[k1]))
            raise RuntimeError
    # @ points
    for k in m1.point_data:
        if verbose > 0:
            print("compare {} in point_data".format(k))
        if not  np.allclose(m1.point_data[k][trace_point_m1_to_m2], m2.point_data[k]):
            print("Diff in {} in point_data".format(k))
            if verbose > 0:
                print(m1.point_data[k][trace_point_m1_to_m2]- m2.point_data[k])
                print(np.linalg.norm(m1.point_data[k][trace_point_m1_to_m2]- m2.point_data[k]))

            raise RuntimeError
    # --------------------------------------------------
    return

def build_diff_mesh(m1, m2):
    assert m1.number_of_points == m2.number_of_points
    assert len(m1.point_data) == len(m2.point_data)
    assert m1.number_of_cells == m2.number_of_cells
    assert len(m1.cell_data) == len(m2.cell_data)

    to_return = m1.copy(deep=True)
    to_return.clear_data()

    for k in m1.cell_data:
        if not np.allclose(m1.cell_data[k], m2.cell_data[k]):
            print("Diff in {} in cell_data".format(k))
            to_return.cell_data[k] = m1.cell_data[k]
            to_return.cell_data[k+"_diff"] = m1.cell_data[k] - m2.cell_data[k]
    # @ points
    for k in m1.point_data:
        if not np.allclose(m1.point_data[k], m2.point_data[k]):
            print("Diff in {} in point_data".format(k))
            to_return.point_data[k] = m1.point_data[k]
            to_return.point_data[k+"_diff"] = m1.point_data[k] - m2.point_data[k]
    return to_return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-11-09 17:39:32.284214"
__version__ = 1.0

from costo.clock import clock as Clock
import traceback

def sould_not_be_called():
    raise RuntimeError

def test_rest_behavior():
    clock = Clock(t0=0)
    clock.reset(0.)
    assert clock.time == 0.

    clock = Clock(t0=0)
    clock.reset(10.)
    assert clock.time == 10.

def test_time_attribute_behavior():
    clock = Clock(t0=0)
    clock.reset(1.)
    # try to set protected atribut time
    try:
        clock.time = 2
        sould_not_be_called()
    except:
        print()
        print("-"*50)
        print('Dont worry, the following exception is a normal behavior')
        print("-"*50)
        ret = traceback.format_exc()
        assert "RuntimeError: Time can't be set, use reset or increment" in ret
        print(ret)
        print("-"*50)
        pass
    assert clock.time == 1.

    # try to acces to an non existing attribut (times)
    try:
        clock.times = 2
        sould_not_be_called()
    except:
        print()
        print("-"*50)
        print('Dont worry, the following exception is a normal behavior')
        print("-"*50)
        ret = traceback.format_exc()
        assert "AttributeError: 'clock' object has no attribute 'times'" in ret
        print(ret)
        print("-"*50)
        pass


def test_tic_tac_behavior_and_next_time_attribut():
    clock = Clock(t0=0)
    clock.reset(0.)

    clock.tic(2.)
    assert clock.next_time == 2.
    clock.tac()
    assert clock.time == 2.
    assert clock.next_time == 2.
    clock.tic(2.)
    assert clock.next_time == 4.
    clock.tac()
    assert clock.next_time == 4.
    assert clock.time == 4.
    clock.reset(1.)
    assert clock.time == 1.
    assert clock.next_time == 1.
    clock.tic(2.)
    assert clock.next_time == 3.
    clock.tac()
    assert clock.next_time == 3.
    assert clock.time == 3.

    # try to acces to an non existing attribut (times)
    try:
        clock.next_time = 12.
        sould_not_be_called()
    except:
        print()
        print("-"*50)
        print('Dont worry, the following exception is a normal behavior')
        print("-"*50)
        ret = traceback.format_exc()
        assert "RuntimeError: next_time can't be set" in ret
        print(ret)
        print("-"*50)
        pass
    assert clock.next_time == 3.

def test_time_singleton_behavior():
    clock = Clock(t0=0)
    clock.reset(3)

    clock2 = Clock(t0=0)
    assert id(clock) == id(clock2)
    assert clock2.time == 3

    clock2.reset(0)
    assert clock.time == 0

    clock.tic(2.5)
    clock.tac()
    assert clock.time == clock2.time
    assert clock2.time == 2.5

    clock2.tic(2.5)
    clock2.tac()
    assert clock.time == clock2.time
    assert clock.time == 5.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# **************************************************
__author__  = "Teddy Chantrait"
__email__   = "teddy.chantrait@gmail.com"
__status__  = "Development"
__date__    = "2024-06-07 18:20:10.418626"
__version__ = "@COSTO_VERSION@"
# **************************************************
import pyvista as pv
from pymetis._internal import part_mesh,GType
from pymetis import MeshPartition, Options
import numpy as np
import inspect
from scipy.spatial import KDTree
from . import utils as COUT


class meshIntersector(pv.UnstructuredGrid, COUT.utils):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self._kdt_elt_center: KDTree = None
        self._kdt_pts: KDTree = None

    def _build_kdt_elt_center(self):
        self._kdt_elt_center = KDTree(self.cell_centers().points)

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def _build_kdt_pts(self):
        self._kdt_pts = KDTree(self.points)

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def build_trace_at_elt_center(self, msh:pv.UnstructuredGrid=None, eps=1e-4,
                                  intersection_field_name=""):
        ifname = intersection_field_name
        self._build_kdt_elt_center()
        dd, ii = self._kdt_elt_center.query(msh.cell_centers().points, eps=eps)
        trace_in_given_mesh = np.argwhere(dd < 1e-7).flatten()
        trace_in_current_mesh = ii[trace_in_given_mesh]
        if ifname != "":
            c_keys = list(msh.cell_data.keys())
            if ifname in c_keys:
                print("WARNING intersection_field_name {} already exist in the \
                given mesh\nKnown cell data are {}".format(ifname, c_keys))
            tmp = np.ones(msh.number_of_cells, dtype=np.int32) * -1
            tmp[ifname] = 1
            msh.interface.cell_data[ifname] = tmp
        return trace_in_given_mesh, trace_in_current_mesh
        # for ifiled in fields_to_transfert["cell"]:
        #     test = np.ones(interface.number_of_cells, dtype=np.int32) * -1
        #     test[trace] = ifiled[1][ii]
        #     interface.cell_data[ifiled[0]] = test

    def build_trace_at_pts(self, pts, eps=1e-4):
        self._build_kdt_pts()
        dd, ii = self._kdt_pts.query(pts, eps=eps)
        trace_in_given_mesh = np.argwhere(dd < 1e-7).flatten()
        trace_in_current_mesh = ii[trace_in_given_mesh]
        return trace_in_given_mesh, trace_in_current_mesh


def build_map_eltype_dim(mesh: pv.UnstructuredGrid=None, verbose=False):
    """
    For each element type of the given mesh, a dictionary is build.It contain two fields:

       - dim (integer): the geometrical dimension of the associated element.
       - vtktype (integer): the associated vtktype

    All dictionaries are retrun in another one that contains the vtk element
    name as key and it associated dictionary.

    Inputs:
       - mesh (pv.UnstructuredGrid): a pyvista unstructured mesh
       - verbose             (bool): active/desavtive the verbosity.

    output:
       - dictionary describe above structured as follow

       {'ELEMENT_TYPE1': {'dim': x},{'vtktype': xx}, 'ELEMENT_TYPE2': {'dim': x}, {'vtktype': xx}, ..., }

    example:

    If a vtk mesh contains lines and triangles, the method return
    the following dictionary:

    {'LINE': {'vtktype': 3, 'dim': 1}, 'TRIANGLE': {'vtktype': 5, 'dim': 2}}
    """
    dims = {}
    uctypes, index = np.unique(mesh.celltypes, return_index=True)

    for vtk_type, idx in zip(uctypes, index):
        cell = mesh.get_cell(idx)
        name = pv.CellType(vtk_type).name
        dim = cell.dimension
        dims[name] = {"vtktype":vtk_type, "dim":dim}
        if verbose:
            print("element: {} (vtktype: {}), space dimension: {}".format(name, vtk_type, dim))
    return dims


def build_bulk_and_interface_elt_fields(omega: pv.UnstructuredGrid=None,
                                        bulk_elt_name:str = "bulk_elt",
                                        interface_elt_name:str = "interface_elt"):
    """
    Generate two (cell_data) fields in the given mesh:
       - for the bulk elements (dim = dim_max)
       - for the interface elements (dim<=dim_max-1)

    Note: the filter is based on the highest geometrical dimension dim_max (of the mesh's elements).

    Inputs:
       - omega (pv.UnstructuredGrid): a pyvista unstructured mesh
       - bulk_elt_name      (string): name of the bulk field generated (default: bulk_elt)
       - interface_elt_name (string): name of the interface field generated (default: interface_elt)

    """
    # END DOC
    c_keys = list(omega.cell_data.keys())
    if bulk_elt_name in c_keys:
        raise NotImplementedError("bulk_elt_name {} already exist in the \
        given mesh\nKnown cell data are {}".format(bulk_elt_name, c_keys))
    if interface_elt_name in c_keys:
        raise NotImplementedError("interface_elt_name {} already exist in the \
        given mesh\nKnown cell data are {}".format(interface_elt_name, c_keys))

    elt_dims = build_map_eltype_dim(omega)
    dims = []
    for val in elt_dims.values():
        dims.append(val["dim"])
    bulk_dim = max(dims)
    elts = list(elt_dims.keys())
    elts.sort()

    assert bulk_dim > 0

    interface_dim = bulk_dim-1
    if elts in (['LINE', 'VERTEX'], ['LINE']):
        print("1D mesh detected")
        bulk_dim = 1
        interface_dim = 3

    bulk_elt = np.zeros(omega.number_of_cells, dtype=np.int32)
    interf_elt = np.zeros(omega.number_of_cells, dtype=np.int32)
    celltypes = omega.celltypes
    for val in elt_dims.values():
        if val["dim"] == bulk_dim:
            trace = np.argwhere(celltypes == val["vtktype"])
            bulk_elt[trace] = 1
        if val["dim"] == interface_dim:
            trace = np.argwhere(celltypes == val["vtktype"])
            interf_elt[trace] = 1

    omega.cell_data[bulk_elt_name] = bulk_elt
    omega.cell_data[interface_elt_name] = interf_elt
    return


def build_graph_partitionning(omega:pv.UnstructuredGrid = None,
                              nb_part=2,
                              part_name="part",
                              part_name_at_node=None):
    """
    Generate a field in the given mesh based on metis
    the field contains an integer related to its partition number for each elements.

    Inputs:
       - omega (pv.UnstructuredGrid): a pyvista unstructured mesh
       - nb_part               (int): number of partition (default: 2)
       - part_name          (string): name of the associated field generated (default: part)
       - part_name_at_node  (string): name of the associated field generated (default: None)

    """
    assert nb_part > 1
    assert type(omega) == pv.UnstructuredGrid
    uniquecelltypes = np.unique(omega.celltypes)
    if len(uniquecelltypes) > 1:
        raise NotImplementedError("heterogenous mesh not yet implemented")

    # --------------------------------------------------
    # use metis to do a graph partitionning
    # --------------------------------------------------
    conn_offset = omega.offset
    conn =  omega.cell_connectivity
    tpwgts = []
    gtype = GType.DUAL
    n_elements = omega.number_of_cells
    n_vertex = omega.number_of_points
    ncommon = 1
    options = Options()
    n_cuts, elem_part, vert_part = MeshPartition(*part_mesh(nb_part, conn_offset, conn,
                                                            tpwgts, gtype, n_elements, n_vertex, ncommon, options))

    # --------------------------------------------------
    omega.cell_data[part_name] = elem_part
    if part_name_at_node:
        omega.point_data[part_name_at_node] = vert_part
    return


def extract_subdomains(omega:pv.UnstructuredGrid = None,
                       part_name="epart"):
    """
    make subdomaines from the given mesh based on the cell_data field part_name.

    Inputs:
        - omega (pv.UnstructuredGrid): a pyvista unstructured mesh
        - part_name          (string): name of the associated integer field to generate subdomaines

    output:
        - the list of the subdomaines generated.
    """
    assert type(omega) == pv.UnstructuredGrid
    c_keys = list(omega.cell_data.keys())
    if part_name not in c_keys:
        raise RuntimeError("field {} (part_name attribute) is not in the mesh (attribute omega)\n\
        known fields ares:\n{}".format(part_name, c_keys))
    parts = np.unique(omega.cell_data[part_name])

    eps = 1e-7
    omegas = []

    for ipart in parts:
        omega_p = omega.threshold([ipart - eps, ipart + eps],
                                  scalars=part_name)
        omegas.append(omega_p)
    return omegas


def map_data_on_local_BC(kdt, interface, fields_to_transfert: list=[]):
    pts = interface.cell_centers().points

    dd, ii = kdt.query(pts, eps=1e-3)
    # print(kdt.n)
    # print(dd)
    # print(pts.shape[0])
    # print(ii)
    # trace = np.argwhere( ii != kdt.n).flatten()
    trace = np.argwhere(dd < 1e-7).flatten()
    ii = ii[trace]
    for ifiled in fields_to_transfert["cell"]:
        test = np.ones(interface.number_of_cells, dtype=np.int32) * -1
        test[trace] = ifiled[1][ii]
        interface.cell_data[ifiled[0]] = test


class Imesh(pv.UnstructuredGrid, COUT.utils):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.msh_dim = None
        self.interface = None
        self.interface_pts = None
        self._field_ids_name = "TMP_CONNECTOR_IDS"
        # self._bb_center = None
        # self._bound = None
        self._init_mesh()

    def _init_mesh(self):
        self._compute_mesh_dim()

    def _compute_mesh_dim(self):
        elt_dims = build_map_eltype_dim(self)

        dims = []
        for elt, val in elt_dims.items():
            dims.append(val["dim"])
        dims = np.unique(np.array(dims, dtype=np.int32))
        if (len(dims) != 1): raise NotImplementedError("Mesh contains element with different dimension")
        self.msh_dim = dims[0]

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def add_local_fields(self, name="connector_ids"):
        """
        add a unique point filed (ordered by point array)
        add a unique cell data field (ordered by cell array)

        Input:
           -name  (string): name of the added field in point and cell data
        """
        c_keys = self.cell_data.keys()
        p_keys = self.point_data.keys()
        if (name in c_keys) or (name in p_keys):
            raise NotImplementedError("field name {} already exist in the mesh. Known fields are\n:@point:\n{}\n@cell:\n{}".format(name, p_keys, c_keys))
        self.point_data[name] = np.arange(self.number_of_points, dtype=np.int32)
        self.cell_data[name] = np.arange(self.number_of_cells, dtype=np.int32)

    def build_skin(self):
        """
        build the interface of the mesh
        """
        self.add_local_fields(self._field_ids_name)
        match self.msh_dim:
            case 1:
                conn = self.cells_dict[pv.CellType.LINE].flatten()
                unique_pts_ids, count = np.unique(conn, return_counts=True)
                trace = np.argwhere(count == 1).flatten()
                # self.interface = self.extract_points(ind=trace, include_cells=False)
                self.interface = pv.PolyData(self.points[trace]).cast_to_unstructured_grid()
                self.interface[self._field_ids_name] = self.point_data[self._field_ids_name][trace]
            case 2:
                surf = self.extract_surface(algorithm='dataset_surface')
                self.interface = surf.extract_feature_edges()
            case 3:
                self.interface = self.extract_surface(algorithm='dataset_surface')
            case _:
                raise RuntimeError
        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing

    def compute_interface_cell_centers(self, ):
        """
        build the interface cell center of the mesh.
        """
        self.build_skin()
        # self.bounds = self.intef_cell_centers.bounds
        # TO DO voir sir il y a moyen de faire un premier try avoir la bounding_box
        self.interface_pts = self.interface.cell_centers().points

        # mecanisme to call only once this fonction
        self_fct_name = inspect.currentframe().f_code.co_name
        self.__dict__[self_fct_name] = self.do_nothing
